(********************************************************************)
(*			   GRAPHIX TOOLBOX 4.0			    *)
(*	   Copyright (c) 1985, 87 Borland International, Inc.	    *)
(*								    *)
(*	   Module graphique pour la carte monochrome Hercules	    *)
(********************************************************************)

unit GDriver;

interface

{$I Float.inc}	{ Dtermine le type Float. }

uses
  Dos, Crt;

{$IFOPT N+}
type
  Float = Double; { Rel sur 8 octets, seulement avec coprocesseur 8087 }

{$ELSE}
type
  Float = real;   { Rel sur 6 octets, aucun coprocesseur ncessaire }

{$ENDIF}

const
  MaxWorldsGlb = 4;
  MaxWindowsGlb = 16;
  MaxPiesGlb = 10;
  MaxPlotGlb = 100;
  StringSizeGlb = 80;
  HeaderSizeGlb = 10;
  RamScreenGlb : boolean = true;
  CharFile : string[StringSizeGlb] = '4x6.fon';
  MaxProcsGlb = 27;
  MaxErrsGlb = 7;
  AspectFactor	= 0.6667;	  { Ratio d'aspect d'un vrai cercle	    }
  ScreenSizeGlb = 16383;	  { Taille totale de l'cran en entiers     }
  HardwareGrafBase : word = $B000;{ position de l'cran matriel            }
  XMaxGlb	= 89;		  { Nombre d'octets -1 dans une ligne cran }
  XScreenMaxGlb = 719;		  { Nombre de pixels -1 dans une ligne cran}
  YMaxGlb	= 349;		  { Nombre de lignes de l'cran -1          }
  IVStepGlb	= 5;		  { Valeur initiale de VStepGlb		    }
  VRowsGlb	= $58;		  { Remplacer par $57 si le moniteur perd
				    sa stabilit horizontale		    }
  MinForeground : word = 1;	  { Valeur minimum de la couleur d'criture }
  MaxForeground : word = 1;	  { Valeur maximum de la couleur d'criture }
  MinBackground : word = 0;	  { Valeur minimum de la couleur de fond    }
  MaxBackground : word = 0;	  { Valeur maximum de la couleur de fond    }
  RamScreenInCard  : boolean = false;
  { Hercules: stocke l'cran virtuel dans le bloc de rserve HGC ?
    Interfrera avec certaines cartes couleurs si prsent. Voir manuel
    Hercules (initialise mode 'Full') }


type
  WrkString = string[StringSizeGlb];
  WrkStringPtr = ^WrkString;
  WorldType = record
		X1, Y1, X2, Y2 : Float;
	      end;
  WindowType = record
		 X1, Y1, X2, Y2 : integer;
		 Header : WrkString;
		 Drawn, Top : boolean;
		 Taille : word;
	       end;
  Worlds = array[1..MaxWorldsGlb] of WorldType;
  Windows = array[1..MaxWindowsGlb] of WindowType;
  PlotArray = array[1..MaxPlotGlb, 1..2] of Float;
  Character = array[1..3] of byte;
  CharArray = array[32..126] of character;
  PieType = record
	      Zone : Float;
	      Text : WrkString;
	    end;
  PieArray = array[1..MaxPiesGlb] of PieType;
  BackgroundArray = array[0..7] of byte;
  LineStyleArray = array[0..7] of boolean;
  ScreenType	    = array[0..ScreenSizeGlb] of word;
  ScreenPointer     = ^ScreenType;
  WindowStackRecord = record
			W : WindowType;
			Contents : ScreenPointer;
		      end;
  Stacks	    = array[1..MaxWindowsGlb] of WindowStackRecord;

var
  X1WldGlb, X2WldGlb, Y1WldGlb, Y2WldGlb, AxGlb, AyGlb, BxGlb, ByGlb : Float;
  X1RefGlb, X2RefGlb, Y1RefGlb, Y2RefGlb : integer;
  LinestyleGlb, MaxWorldGlb, MaxWindowGlb, WindowNdxGlb, WorldNdxGlb : integer;
  X1Glb, X2Glb, Y1Glb, Y2Glb : integer;
  XTextGlb, YTextGlb, VStepGlb : integer;
  PieGlb, DirectModeGlb, ClippingGlb, AxisGlb, HatchGlb : boolean;
  MessageGlb, BrkGlb, HeaderGlb, TopGlb, GrafModeGlb : boolean;
  CntGlb, ColorGlb : byte;
  ErrCodeGlb : integer;
  LineStyleArrayGlb : LineStyleArray;
  ErrorProc : array[0..MaxProcsGlb] of WrkStringPtr;
  ErrorCode : array[0..MaxErrsGlb] of WrkStringPtr;
  PcGlb : string[40];
  AspectGlb : Float;
  GrafBase : word;
  World : Worlds;
  GrafWindow : Windows;
  CharSet : CharArray;
  ScreenGlb : ScreenPointer;
  Stack : Stacks;

function BaseAddress(Y : word) : word;
{ Calcule l'adresse de la ligne Y }

procedure Error(ProcErr, CodeErr : integer);

function HardwarePresent : boolean;
{ Teste la prsence d'une carte graphique }

procedure HerculesMode(Mode : word); { 0 = Texte, 1 = Graphique }

procedure AllocateRAMScreen;
{ Alloue l'cran virtuel et s'assure que ScreenGlb
  se situe sur une frontire de segment (16 octets) }

procedure LeaveGraphic;
{ Sort du mode graphique et efface l'cran }

procedure DC(C : byte);
{ Dessine le caractre avec le code ASCII C
   la position (XTextGlb, YTextGlb). }

procedure SetForegroundColor(Couleur : word);

procedure SetBackgroundColor(Couleur : word);

procedure ClearScreen;
{ Efface l'cran affich }

procedure EnterGraphic;
{ Active mode graphique }

procedure DP(X, Y : word);
{ Allume un pixel en (X, Y). }

function PD(X, Y : word) : boolean;
{ Retourne True si la couleur du pixel en (X, Y) correspond  ColorGlb }

procedure SetBackground8(Fond : BackgroundArray);
{ Remplit l'affichage actif avec la squence de bits spcifie }

procedure SetBackground(Oct : byte);
{ Dtermine le fond de la fentre active }

procedure DrawStraight(X1, X2, Y : word);
{ Dessine une ligne horizontale de X1,Y  X2,Y }

procedure SaveScreen(NomFich : WrkString);
{ Sauvegarde l'cran courant sur le disque sous le nom NomFich. }

procedure LoadScreen(NomFich : WrkString);
{ Charge l'cran depuis le fichier NomFich. }

procedure SwapScreen;
{ Echanges le contenu de l'cran affich avec le contenu de l'cran virtuel }

procedure CopyScreen;
{ Copie l'cran actif sur l'cran inactif }

procedure InvertScreen;
{ Inverse l'image de l'cran actif }

implementation

const
  FontLoaded	     : boolean = false; { La fonte est-elle charge ? }
  ForegroundColorGlb : word = 1;

type
  FontChar = array[0..13] of byte;
  GrfFont  = array[0..255] of FontChar;

var
  Font : GrfFont;

function BaseAddress{(Y : word) : word};
{ Calcule l'adresse de la ligne Y }
begin
  BaseAddress := (Y and 3) shl 13 + 90 * (Y shr 2);
end; { BaseAddress }

procedure Error{(ProcErr, CodeErr : integer)};
var
  XLoc, YLoc : integer;
  Ch : char;

begin { Error }
  if not (ProcErr in [0..MaxProcsGlb]) then
  begin
    LeaveGraphic;
    WriteLn('Erreur FATALE 1: Procdure interdite numro ', ProcErr);
    Halt;
  end;
  if not (CodeErr in [0..MaxErrsGlb]) then
  begin
    LeaveGraphic;
    WriteLn('Erreur FATALE 2: Code erreur interdit ', CodeErr);
    Halt;
  end;
  ErrCodeGlb := CodeErr;
  if BrkGlb then
    LeaveGraphic;
  if MessageGlb or BrkGlb then
  begin
    XLoc := XTextGlb;
    YLoc := YTextGlb;
    GotoXY(1, 24);
    ClrEOL;
    WriteLn('Erreur Turbo Graphix No ', CodeErr, ' dans procdure No ', ProcErr);
    if MessageGlb then
    begin
      ClrEOL;
      Write('(', ErrorCode[CodeErr]^, ' dans ', ErrorProc[ProcErr]^, ')');
    end;
  end;
  if BrkGlb then
    Halt
  else if MessageGlb then
    begin
      Write('.  Pressez enter: ');
      repeat
	Ch := ReadKey;
      until (Ch = ^M) or (Ch = ^C);
      if Ch = ^C then
      begin
	LeaveGraphic;
	Halt;
      end;
      GotoXY(XLoc, YLoc);
    end;
end; { Error }

function HardwarePresent{ : boolean};
{ Teste la prsence d'une carte graphique }
var
  Regs : Registers;

function HercPresent : boolean;
begin
  inline($BB/$00/$01/	       {      MOV    BX,0100H		  }
	 $BA/$BA/$03/	       {      MOV    DX,03BAH		  }
	 $EC/		       {      IN     AL,DX		  }
	 $88/$C4/	       {      MOV    AH,AL		  }
	 $80/$E4/$80/	       {      AND    AH,80H		  }
	 $B9/$40/$00/	       { L3:  MOV    CX,0040H		  }
	 $EC/		       { L1:  IN     AL,DX		  }
	 $24/$80/	       {      AND    AL,80H		  }
	 $38/$E0/	       {      CMP    AL,AH		  }
	 $E1/$F9/	       {      LOOPZ  L1			  }
	 $75/$05/	       {      JNZ    L2			  }
	 $4B/		       {      DEC    BX			  }
	 $75/$F1/	       {      JNZ    L3			  }
	 $EB/$33/	       {      JMP    L4			  }
	 $B8/$00/$B0/	       { L2:  MOV    AX,B000H		  }
	 $8E/$C0/	       {      MOV    ES,AX		  }
	 $E8/$11/$00/	       {      CALL   L5			  }
	 $75/$0B/	       {      JNZ    L6			  }
	 $B0/$01/	       {      MOV    AL,01		  }
	 $BA/$BF/$03/	       {      MOV    DX,03BFH		  }
	 $EE/		       {      OUT    DX,AL		  }
	 $E8/$06/$00/	       {      CALL   L5			  }
	 $74/$1E/	       {      JZ     L4			  }
	 $B0/$01/	       { L6:  MOV    AL,01		  }
	 $EB/$1C/	       {      JMP    L7			  }
	 $26/$8A/$1E/$FF/$7F/  { L5:  MOV    BL,ES:[7FFFH]	  }
	 $26/$8A/$0E/$FF/$3F/  {      MOV    CL,ES:[3FFFH]	  }
	 $26/$FE/$06/$FF/$3F/  {      INC    BYTE PTR ES:[3FFFH]  }
	 $26/$3A/$1E/$FF/$3F/  {      CMP    BL,ES:[3FFFH]	  }
	 $26/$88/$0E/$FF/$3F/  {      MOV    ES:[3FFFH],CL	  }
	 $C3/		       {      RET			  }
	 $30/$C0);	       { L4:  XOR    AL,AL		  }
			       { L7:				  }

end; { HercPresent }

begin
  Intr($11, Regs);
  HardwarePresent := HercPresent and ((Regs.AX and 48) = 48);
end; { HardwarePresent }

procedure HerculesMode{(Mode : word)}; { 0=texte, 1=graphique }
type
  ModeDescriptor = record
		     CRTMode : byte;
		     R6845   : array[0..11] of byte;
		   end;
const
  HercModes : array[0..1] of ModeDescriptor =
  ((CRTMode : 32; R6845 : ($61,$50,$52,$0F,$19,$06,$19,$19,$02,$0D,$0B,$0C)),
   (CRTMode : 2;  R6845 : ($35,$2D,$2E,$07,$5B,$02,$58,$58,$02,$03,$00,$00)));
  ScreenOn	    = 8;
  IndexPort	    = $3B4;
  DataPort	    = $3B5;
  ControlPort	    = $3B8;
  ConfigurationPort = $3BF;

var
  I : word;

begin
  Port[ConfigurationPort] := 1;
  Port[ControlPort] := HercModes[Mode].CRTMode;
  for I := 0 to 11 do
  begin
    Port[IndexPort] := I;
    Port[DataPort] := HercModes[Mode].R6845[I];
  end;
  Port[ControlPort] := HercModes[Mode].CRTMode or ScreenOn;
  if RAMScreenInCard then
    Port[ConfigurationPort] := 3;
end; { HerculesMode }

procedure AllocateRAMScreen;
{ Alloue l'cran virtuel et s'assure que ScreenGlb
  se situe sur une frontire de segment (16 octets) }
var
  Test : ^byte;
begin
  if RamScreenInCard then
    ScreenGlb := Ptr($B800, $0000)
  else
    begin
      New(ScreenGlb);
      while Ofs(ScreenGlb^) <> 0 do
      begin
	Dispose(ScreenGlb);
	New(Test);
	New(ScreenGlb);
      end;
    end;
end; { AllocateRAMScreen }

{$L GrafHGC.OBJ}
procedure DC{(C : byte)}; external;

procedure DP{(X, Y : word)}; external;

procedure SwapScreen; external;

procedure InvertScreen; external;

{$F+}
function WriteGrafChars(var F : TextRec) : integer;
{ Utile pour sortir des caractres graphiques
  par le canal de sortie standard }
const
  BackSpace = #8;
  LineFeed  = #10;
  Return    = #13;
var
  I : integer;
begin
  with F do
    if Mode = fmOutput then
    begin
      if BufPos > BufEnd then
      begin
	for I := BufEnd to Pred(BufPos) do  { Vide le buffer de sortie }
	begin
	  case BufPtr^[I] of
	    BackSpace : if XTextGlb > 1 then
			  DEC(XTextGlb);

	    LineFeed  : if YTextGlb < 25 then
			  INC(YTextGlb);

	    Return    : XTextGlb := 1;
	  else
	    DC(ORD(BufPtr^[I]));
	    if XTextGlb < 80 then
	      INC(XTextGlb);
	  end; { case }
	end; { for }
      end;
      BufPos := BufEnd;
    end; { if }
  WriteGrafChars := 0;
end; { WriteGrafChars }

function GrafCharZero(var F : TextRec) : integer;
{ Appele lorsque la sortie standard est ouverte et referme }
begin
  GrafCharZero := 0;
end; { GrafCharZero }
{$F-}

var
  OldOutput : Text; {Stocke le canal E/S de sortie }

procedure GrafCharsON;
{ Redirige la sortie standard vers la fonction WriteGrafChars. }
begin
  Move(Output, OldOutput, SizeOf(Output));  { Sauve ancien canal de sortie }
  with TextRec(Output) do
  begin
    OpenFunc:=@GrafCharZero;	   { pas d'ouverture ncessaire }
    InOutFunc:=@WriteGrafChars;    { WriteGrafChars appele pour E/S }
    FlushFunc:=@WriteGrafChars;    { WriteGrafChars se vide automatiquement }
    CloseFunc:=@GrafCharZero;	   { pas de fermeture ncessaire }
    Name[0]:=#0;
  end;
end; { GrafCharsON }

procedure GrafCharsOFF;
{ Restaure le canal E/S de sortie d'origine }
begin
  Move(OldOutput, Output, SizeOf(OldOutput));
end; { GrafCharsOFF }

procedure LeaveGraphic;
{ Sort du mode graphique et efface l'cran }
var
  Regs : Registers;
begin
  HerculesMode(0);
  ClrScr;
  GrafCharsOFF;
  GrafModeGlb := false;
end; { LeaveGraphic }

procedure SetForegroundColor{(Couleur : word)};
begin
  { Pas de couleurs  choisir }
end; { SetForegroundColor }

procedure SetBackgroundColor{(Couleur : word)};
begin
  { Pas de couleurs  choisir }
end; { SetBackgroundColor }

procedure ClearScreen;
{ Efface l'cran affich }
begin
  FillChar(Mem[GrafBase:0000], ScreenSizeGlb shl 1, 0);
end; { ClearScreen }

procedure EnterGraphic;
{ Active le mode Enter graphique }
var
  FontFile : file of GrfFont;
begin
  if not FontLoaded then
  begin
    Assign(FontFile, '14x9.FON');
    {$I-} Reset(FontFile); {$I+}
    if IOresult=0 then
      begin
	Read(FontFile, Font);
	Close(FontFile);
      end
    else
      FillChar(Font, SizeOf(Font), 0);
    FontLoaded := true;
  end;
  HerculesMode(1);
  SetForegroundColor(MaxForeground);
  if not GrafModeGlb then
   GrafCharsON;
  ClearScreen;
  GrafModeGlb := true;
end; { EnterGraphic }

function PD{(X, Y : word) : boolean};
{ Retourne True si la couleur du pixel en (X, Y) correspond  ColorGlb }
begin
  PD := (ColorGlb = 0) xor (Mem[GrafBase:BaseAddress(Y) + X shr 3]
		       and (128 shr (X and 7)) <> 0);
end; { PD }

procedure SetBackground8{(Fond : BackgroundArray)};
{ Remplit l'affichage actif avec la squence de bits spcifie }
var
  I : word;
begin
  for I := Y1RefGlb to Y2RefGlb do
    FillChar(Mem[GrafBase:BaseAddress(I) + X1RefGlb], X2RefGlb - X1RefGlb+1,
	     Fond[I and 7]);
end; { SetBackground8 }

procedure SetBackground{(Oct : byte)};
{ Dtermine le fond de la fentre active }
var
  Bk : BackgroundArray;
begin
  FillChar(Bk, 8, Oct);
  SetBackground8(Bk);
end; { SetBackground }

procedure DrawStraight{(X1, X2, Y : word)};
{ Dessine une ligne horizontale de X1,Y  X2,Y }
var
  I, X : word;
  PosModeDirect : boolean;
begin
  if (not ((X1 < 0) or (X1 > XMaxGlb shl 3+7)) and not ((X2 < 0) or
     (X2 > XMaxGlb shl 3+7)) and ((Y >= 0) and (Y <= YMaxGlb))) then
  begin
    PosModeDirect := DirectModeGlb;
    DirectModeGlb := true;
    if X1 > X2 then
    begin
      X := X1;
      X1 := X2;
      X2 := X;
    end;
    if X2-X1 < 16 then
      for X := X1 to X2 do
	DP(X, Y)
    else
      begin
	X1 := X1 + 8;
	for I := (X1 - 8) to (X1 and -8) do
	  DP(I, Y);
	for I := (X2 and -8) to X2 do
	  DP(I, Y);
	FillChar(Mem[GrafBase:BaseAddress(Y) + (X1 shr 3)],
		(X2 shr 3) - (X1 shr 3), ColorGlb);
      end;
    DirectModeGlb := PosModeDirect;
  end
end; { DrawStraight }

procedure SaveScreen{(NomFich : WrkString)};
{ Sauvegarde l'cran courant sur le disque sous le nom NomFich. }
type
  Sector = array[0..127] of byte;
  SecScreen = array[0..255] of sector;
var
  PictureFile : file of Sector;
  Pic : ^SecScreen;
  I : word;
  IOErr : boolean;

procedure IOCheck;
begin
  IOErr := IOResult <> 0;
  if IOErr then
    Error(27, 5);
end; { IOCheck }

begin
  if NomFich <> '' then
    begin
      IOErr := false;
      Pic := Ptr(GrafBase, 0);
      Assign(PictureFile, NomFich);
      {$I-}
      Rewrite(PictureFile);
      {$I+}
      IOCheck;
      for I := 0 to 255 do
	if not IOErr then
	begin
	  {$I-}
	  Write(PictureFile, Pic^[I]);
	  {$I+}
	  IOCheck;
	end;
      if not IOErr then
      begin
	{$I-}
	Close(PictureFile);
	{$I+}
	IOCheck;
      end;
    end
  else
    Error(27, 5);
end; { SaveScreen }

procedure LoadScreen{(NomFich : WrkString)};
{ Charge l'cran depuis le fichier NomFich. }
type
  Sector = array[0..127] of byte;
  SecScreen = array[0..255] of Sector;
var
  PictureFile : file of Sector;
  Pic : ^SecScreen;
  I : word;
begin
  if NomFich <> '' then
    begin
      Pic := Ptr(GrafBase, 0);
      Assign(PictureFile, NomFich);
      {$I-}
      Reset(PictureFile);
      {$I+}
      if IOResult <> 0 then
	Error(11, 5)
      else
	begin
	  for I := 0 to 255 do
	    Read(PictureFile, Pic^[I]);
	  Close(PictureFile);
	end;
    end
  else
    Error(11, 5);
end; { LoadScreen }

procedure CopyScreen;
{ Copie l'cran actif sur l'cran inactif }
var
  ToBase : word;
begin
  if RamScreenGlb then
  begin
    if GrafBase = HardwareGrafBase then
      ToBase := Seg(ScreenGlb^)
    else
      ToBase := HardwareGrafBase;
    Move(Mem[GrafBase:0000], Mem[ToBase:0000], ScreenSizeGlb shl 1);
  end;
end; { CopyScreen }

end. { GDriver }
